#!/usr/bin/env python
#
# test_importall.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import pkgutil
import importlib
import fsleyes_props as props


def test_importall():

    def recurse(module):

        path    = module.__path__
        name    = module.__name__
        submods = list(pkgutil.iter_modules(path, '{}.'.format(name)))

        for i, (spath, smodname, ispkg) in enumerate(submods):

            submod = importlib.import_module(smodname)

            if ispkg:
                recurse(submod)

    recurse(props)



#!/usr/bin/env python
#
# test_cache.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import              gc
import itertools as it

import fsleyes_props as props

import pytest


class MyObj(props.HasProperties):
    myint  = props.Int()
    mybool = props.Boolean(default=False)
    mystr  = props.String()


def test_PropCache():


    count  = [0]
    target = MyObj()

    def keyfunc():
        key = count[0]
        count[0] += 1
        return key

    # values that the target properties
    # will take. The first ones are the
    # initial values
    values = {
        'myint'  : [3, 5, 9],
        'mybool' : [True, False, True],
        'mystr'  : ['abc', '123', 'ABC']
    }

    triggers = [(MyObj(), 'myint'),
                (MyObj(), 'mybool')]

    cache = props.PropCache(
        target,
        ['myint', 'mybool', 'mystr'],
        keyfunc,
        triggers)

    target.myint  = values['myint'][ 0]
    target.mybool = values['mybool'][0]
    target.mystr  = values['mystr'][ 0]

    with pytest.raises(ValueError):
        cache.get(0, 'myint', 'a', 'b')

    # if no values cached, should
    # return the current value
    assert cache.get(999, 'myint')  == values['myint'][ 0]
    assert cache.get(999, 'mybool') == values['mybool'][0]
    assert cache.get(999, 'mystr')  == values['mystr'][ 0]

    # Or a default value if we provide one
    assert cache.get(999, 'myint',  'blabs') == 'blabs'
    assert cache.get(999, 'mybool', 'blabs') == 'blabs'
    assert cache.get(999, 'mystr',  'blabs') == 'blabs'

    # trigger the cache
    triggers[0][0].myint = 99

    # change the current values
    target.myint  = values['myint'][ 1]
    target.mybool = values['mybool'][1]
    target.mystr  = values['mystr'][ 1]

    # check the cache
    assert cache.get(0, 'myint')  == values['myint'][ 0]
    assert cache.get(0, 'mybool') == values['mybool'][0]
    assert cache.get(0, 'mystr')  == values['mystr'][ 0]

    # trigger the cache again
    triggers[1][0].mybool = True

    # change the current values
    target.myint  = values['myint'][ 2]
    target.mybool = values['mybool'][2]
    target.mystr  = values['mystr'][ 2]

    # check the cache
    for i in range(2):

        assert cache.get(i, 'myint')  == values['myint'][ i]
        assert cache.get(i, 'mybool') == values['mybool'][i]
        assert cache.get(i, 'mystr')  == values['mystr'][ i]


def test_PropCache_gc():

    clearThings = ['target', 'keyfunc']

    for clearThing in clearThings:
        _test_PropCache_gc(clearThing)


def _test_PropCache_gc(clearThing):

    count  = [0]
    target = MyObj()

    def keyfunc():
        key = count[0]
        count[0] += 1
        return key

    triggers = [(MyObj(), 'myint'),
                (MyObj(), 'mybool')]

    cache = props.PropCache(target,
                            ['myint',
                             'mybool',
                             'mystr'],
                            keyfunc,
                            triggers)

    target.myint  = 5
    target.mybool = True
    target.mystr  = 'abc'

    # Put something in the cache
    triggers[0][0].myint = 99

    target.myint  = 99
    target.mybool = False
    target.mystr  = '123'

    assert cache.get(0, 'myint')  == 5
    assert cache.get(0, 'mybool') is True
    assert cache.get(0, 'mystr')  == 'abc'

    # GC the keyfunc or the target
    if   clearThing == 'target':  del target
    elif clearThing == 'keyfunc': del keyfunc

    gc.enable()
    gc.collect()

    # Test cache retrieval or triggering
    if clearThing == 'target':
        assert not cache.alive()
        with pytest.raises(props.CacheError):
            print(cache.get(0, 'myint'))


    elif clearThing == 'keyfunc':
        triggers[1][0].mybool = True
        assert not cache.alive()
        with pytest.raises(props.CacheError):
            print(cache.get(0, 'myint'))



#!/usr/bin/env python


import fsleyes_props as props


def test_suppress():

    class Foo(props.Props):
        num1 = props.Int()
        num2 = props.Int()

    p = Foo()

    called = {}

    def num1Changed():
        called['num1'] = True
    def num2Changed():
        called['num2'] = True

    p.listen('num1', 'listener', num1Changed)
    p.listen('num2', 'listener', num2Changed)

    p.num1 = 1
    assert called == {'num1' : True}
    p.num2 = 1
    assert called == {'num1' : True, 'num2' : True}

    called = {}

    with props.suppress(p, 'num1'):
        p.num1 = 4
        p.num2 = 4
    assert called == {'num2' : True}

    called = {}
    with props.suppress(p, 'num1', 'num2'):
        p.num1 = 6
        p.num2 = 6
    assert called == {}

    with props.suppressAll(p):
        p.num1 = 7
        p.num2 = 7
    assert called == {}

    p.num1 = 8
    p.num2 = 8
    assert called == {'num1' : True, 'num2' : True}


def test_skip():

    class Foo(props.Props):
        num1 = props.Int()
        num2 = props.Int()

    p = Foo()

    called = {}

    def num1Changed1():
        called['num11'] = True
    def num2Changed1():
        called['num21'] = True

    def num1Changed2():
        called['num12'] = True
    def num2Changed2():
        called['num22'] = True

    p.listen('num1', 'listener1', num1Changed1)
    p.listen('num2', 'listener1', num2Changed1)
    p.listen('num1', 'listener2', num1Changed2)
    p.listen('num2', 'listener2', num2Changed2)

    p.num1 = 1
    p.num2 = 1
    assert called == {'num11' : True, 'num21' : True,
                      'num12' : True, 'num22' : True}

    called = {}
    with props.skip(p, 'num1', 'listener1'):
        p.num1 = 2
        p.num2 = 2
    assert called == {'num12' : True, 'num21' : True, 'num22' : True}

    called = {}
    with props.skip(p, 'num2', 'listener1'):
        p.num1 = 3
        p.num2 = 3
    assert called == {'num11' : True, 'num12' : True, 'num22' : True}

    called = {}
    with props.skip(p, ('num1', 'num2'), 'listener1'):
        p.num1 = 4
        p.num2 = 4
    assert called == {'num12' : True, 'num22' : True}

    called = {}
    with props.skip(p, ('num1', 'num2'), 'listener2'):
        p.num1 = 5
        p.num2 = 5
    assert called == {'num11' : True, 'num21' : True}



#!/usr/bin/env python
#
# cache.py - The PropCache class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`PropCache` class, a convenience class
for caching the values of the properties of a :class:`.HasProperties` instance.
"""


import weakref

import fsl.utils.weakfuncref as weakfuncref


class CacheError(Exception):
    """Error raised by the :class:`PropCache` when a reference to the target
    object being cached, or the key function, is lost.
    """
    pass


class PropCache:
    """The ``PropCache`` A little convenience class which can be used to
    track and cache the property values of a :class:`.HasProperties` instance.

    Whenever a *trigger* property changes, the ``PropCache`` will take a copy
    of all specified propertyies on the *target* :class:`.HasProperties`
    instance, and store them alongside a *key*, which is generated by a
    function that is supplied.


    The ``PropCache`` maintains weak references to the target object, the
    key function, and the trigger objects. If the target or key function
    are garbage-collected, the ``PropCache`` will effectively disable itself,
    and subsequent calls to :meth:`get` will raise a :exc:`CacheError`.
    """


    def __init__(self, target, propNames, keyFunc, triggers):
        """Create a ``PropCache``.

        :arg target:    Target :class:`.HasProperties` instance whose
                        property values are to be cached.

        :arg propNames: Names of properties on the ``target`` to be cached

        :arg keyFunc:   Function which will be called whenever a trigger
                        property changes, to generate a suitable key to
                        use for cached property values.

        :arg triggers:  Sequence of ``(hasProps, propName)`` pairs, specifying
                        which property changes should trigger caching of
                        the ``target`` property values.
        """

        self.__name      = '{}_{}'.format(type(self).__name__, id(self))
        self.__cache     = {}
        self.__target    = weakref.ref(target)
        self.__propNames = propNames
        self.__keyFunc   = weakfuncref.WeakFunctionRef(keyFunc)
        self.__triggers  = [(weakref.ref(tobj), tprop)
                            for (tobj, tprop) in triggers]

        for tobj, tprop in triggers:
            tobj.addListener(tprop,
                             self.__name,
                             self.__doCache,
                             immediate=True)


    def __deregister(self):
        """Called if the target or key function are garbage-collected.
        De-registers trigger listeners, and removes all references and
        cached property values.
        """

        if self.__target is None:
            return

        for tobj, tprop in self.__triggers:
            tobj = tobj()
            if tobj is not None:
                tobj.removeListener(tprop, self.__name)

        self.__target   = None
        self.__keyFunc  = None
        self.__cache    = None
        self.__triggers = None


    def alive(self):
        """Returns ``True`` if this ``PropCache`` is active, ``False``
        otherwise. A ``PropCache`` will become inactive if its target,
        or the provided key generation function is deleted.
        """
        return self.__cache is not None and self.__target() is not None


    def get(self, key, propName, *args):
        """Returns the cached property value for the specified ``key``.
        If there is no cached property value for the overlay, the specified
        ``default`` value is returned. If a ``default`` value is not
        provided, the current property value is returned.

        :arg overlay:  Key to retrieve the property value for

        :arg propName: Name of the property to return a value for

        :arg default:  Value to return if a value for the
                       key/property is not cached. Must be passed as a
                       positional argument.
        """
        if len(args) not in (0, 1):
            raise ValueError('Invalid arguments passed to PropCache.get')

        haveDefault = len(args) == 1
        target      = self.__target() if self.__target else None

        if target is None:
            self.__deregister()
            raise CacheError('Target has been gc\'d')

        val = self.__cache.get((key, propName), None)

        if val is None:
            if haveDefault: val = args[0]
            else:           val = getattr(target, propName)

        return val


    def __doCache(self, *a):
        """Called when any trigger property changes. Generates a key, and
        then caches property values on the target instance.
        """

        target  = self.__target()  if self.__target  else None
        keyFunc = self.__keyFunc() if self.__keyFunc else None

        if keyFunc is None or target is None:
            self.__deregister()
            return

        key = keyFunc()

        for propName in self.__propNames:
            self.__cache[key, propName] = getattr(target, propName)



#!/usr/bin/env python
#
# serialise.py - Functions for serialising/deserialising property values.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides functions for serialising and deserialising the values
of most ``fsleyes_props`` property types (which are defined in
:mod:`.properties_types`).


This module only contains logic for those property types that the default
python conversion routines would not be able to handle. For example, an
:class:`.Int` property value can be serialised with ``str``, and deserialised
with ``int``. Such logic is already built into the ``Int`` class, so there
is no need for this module to duplicate it.


An example of where specific serialisation/deserialisation logic is useful is
the :class:`.Boolean` property. Passing a boolean value,``True`` or ``False``
to the ``str`` function will result in a string ``'True'`` or
``'False'``. However, passing a string ``'True'`` or ``'False'`` to the
``bool`` function will result in ``True`` in both cases.


The logic needed to safely serialise/deserialise :class:`.Boolean` values, and
other property types for which it is needed, is provided by this module.
"""

import sys
import logging

import numpy as np


log = logging.getLogger(__name__)


DELIMITER = '#'
"""Delimiter string used to join multi-valued properties. """


def serialise(hasProps, propName):
    """Get the value of the named property from the given
    :class:`.HasProperties` instance, serialise it, and return the
    serialised string.
    """

    propObj  = hasProps.getProp(propName)
    val      = getattr(hasProps, propName)
    propType = type(propObj).__name__
    sfunc    = getattr(sys.modules[__name__],
                       '_serialise_{}'.format(propType),
                       None)

    def defaultSfunc(s, *a):
        return str(s)

    if sfunc is None:
        sfunc = defaultSfunc

    sval = sfunc(val, hasProps, propObj)

    log.debug('Serialised {}.{}: {} -> "{}"'.format(
        type(hasProps).__name__, propName, val, sval))

    return sval


def deserialise(hasProps, propName, string):
    """Deserialise the given string, under the assumption that it is a
    serialised value of the named property from the given
    :class:`.HasProperties` instance.
    """

    propObj  = hasProps.getProp(propName)
    propType = type(propObj).__name__
    dfunc    = getattr(sys.modules[__name__],
                       '_deserialise_{}'.format(propType),
                       None)

    def defaultDfunc(s, *a):
        return s

    if dfunc is None:
        dfunc = defaultDfunc

    dval = dfunc(string, hasProps, propObj)

    log.debug('Deserialised {}.{}: "{}" -> {}'.format(
        type(hasProps).__name__, propName, string, dval))

    return dval


# The type specific conversions are performed by the
# functions below. They must accept the following
# arguments:
#
#   - The value to be serialised/deserialised
#   - The HasProperties instance
#   - The PropertyBase instance


def _serialise_Boolean(value, *a):
    return str(value)


def _deserialise_Boolean(value, *a):

    # Special case - a string containig 'false'
    # (case insensitive) evaluates to False.
    if isinstance(value, str):
        value = value.lower()
        if value == 'false':
            value = ''

    # For anything else, we
    # rely on default conversion.
    return bool(value)


def _serialise_Choice(value, *a):
    return str(value)


def _deserialise_Choice(value, hasProps, propObj):

    choices = propObj.getChoices(hasProps)

    # This is a bit hacky - Choice properties can store
    # any type, so we can't figure out precisely how
    # to serialise/deserialise those types. So we check
    # to see if all the choices are numeric and, if not,
    # fall back to using str for deserialisation.
    if   all([isinstance(c,  float) for c in choices]): cType = float
    elif any([isinstance(c,  float) for c in choices]): cType = float
    elif all([isinstance(c,  int)   for c in choices]): cType = int
    else:                                               cType = str

    return cType(value)


def _serialise_Colour(value, *a):

    # Colour values should be in the range [0, 1]
    r, g, b, a = [int(v * 255) for v in value]
    hexstr     = '#{:02x}{:02x}{:02x}{:02x}'.format(r, g, b, a)

    return hexstr


def _deserialise_Colour(value, *a):

    r = value[1:3]
    g = value[3:5]
    b = value[5:7]
    a = value[7:9]

    r, g, b, a = [int(v, base=16) for v in (r, g, b, a)]
    r, g, b, a = [v / 255.0       for v in (r, g, b, a)]

    return [r, g, b, a]


def _serialise_ColourMap(value, *a):
    return value.name


def _deserialise_ColourMap(value, *a):
    import matplotlib.cm as mplcm
    return mplcm.get_cmap(value)


def _serialise_Bounds(value, *a):
    value = map(str, value)
    return DELIMITER.join(value)


def _deserialise_Bounds(value, *a):
    value = value.split(DELIMITER)
    return map(float, value)


def _serialise_Point(value, *a):
    value = map(str, value)
    return DELIMITER.join(value)


def _deserialise_Point(value, *a):
    value = value.split(DELIMITER)
    return map(float, value)


def _serialise_Array(value, *a):
    ndim  = str(value.ndim)
    shape = [str(s) for s in value.shape]
    value = [str(v) for v in value.flat[:]]
    return DELIMITER.join([ndim] + shape + value)


def _deserialise_Array(value, *a):
    value = value.split(DELIMITER)
    ndim  = int(value[0])
    shape = [int(  v) for v in value[1:ndim + 1]]
    value = [float(v) for v in value[  ndim + 1:]]
    return np.array(value).reshape(shape)



#!/usr/bin/env python
#
# properties.py - Python descriptor framework.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""Python descriptor framework.

This module defines the :class:`PropertyBase`, :class:`ListPropertyBase`,
which form the basis for defining class properties; and the
:class:`HasProperties` class, which is intended to be sub-classed by
application code.

 .. autosummary::
    :nosignatures:

    HasProperties
    PropertyBase
    ListPropertyBase
"""

import weakref
import logging
import warnings

from . import properties_value
from . import bindable
from . import serialise


log = logging.getLogger(__name__)


class _InstanceData:
    """An ``_InstanceData`` object is created for every ``PropertyBase``
    object of a ``HasProperties`` instance. It stores references to the
    instance and the associated :class:`.PropertyValue` instance.
    """

    def __init__(self, instance, propVal):
        self.instance = weakref.ref(instance)
        self.propVal  = propVal


class DisabledError(Exception):
    """A ``DisabledError`` is raised when an attempt is made to assign
    a value to a disabled property. See the :meth:`PropertyBase.enable`
    and :meth:`PropertyBase.disable` methods.
    """
    pass


class PropertyBase:
    """The base class for properties.

    For every ``HasProperties`` instance which has this ``PropertyBase``
    object as a property, one ``_InstanceData`` object is created and attached
    as an attribute of the ``HasProperties`` object.

    One important point to note is that a ``PropertyBase`` object may exist
    without being bound to a ``HasProperties`` instance (in which case it will
    not create or manage any ``PropertyValue`` instances). This is useful
    if you just want validation functionality via the :meth:`validate`,
    :meth:`getAttribute` and :meth:`setAttribute` methods, passing in
    ``None`` for the instance parameter. Nothing else will work properly
    though.

    Several subclasses are defined in the :mod:`.properties_types` module.
    All subclasses should:

      - Ensure that the superclass :meth:`__init__` is called.

      - Override the :meth:`validate` method to implement any built in
        validation rules, ensuring that the the superclass implementation
        is called first (see the :class:`.Number` property for an example).

      - Override the :meth:`cast` method for implicit casting/conversion logic
        (see the :class:`.Boolean` property for an example).
    """

    def __init__(self,
                 default=None,
                 validateFunc=None,
                 equalityFunc=None,
                 required=False,
                 allowInvalid=True,
                 **atts):
        """Define a ``PropertyBase`` property.

        :param default:       Default/initial value.

        :param bool required: Boolean determining whether or not this
                              property must have a value. May alternately
                              be a function which accepts one parameter,
                              the owning ``HasProperties`` instance,
                              and returns ``True`` or ``False``.

        :param validateFunc:  Custom validation function. Must accept three
                              parameters: a reference to the ``HasProperties``
                              instance, the owner of this property; a
                              dictionary containing the attributes for this
                              property; and the new property value. Should
                              return ``True`` if the property value is valid,
                              ``False`` otherwise.

        :param equalityFunc:  Function for testing equality of two values.

        :param allowInvalid:  If ``False``, a :exc:`ValueError` will be
                              raised on all attempts to set this property
                              to an invalid value. This does not guarantee
                              that the property value will never be
                              invalid - see caveats in the
                              :class:`.PropertyValue` documentation.

        :param atts:          Type specific attributes used to test
                              validity - passed to the
                              :meth:`.PropertyValue.__init__` method as
                              its ``attributes``.
        """

        atts['default'] = default
        atts['enabled'] = atts.get('enabled', True)

        # A _label is added to this dict by
        # HasProperties.__new__ class when
        # the first instance of that class
        # is created
        self._label             = {}
        self._required          = required
        self._validateFunc      = validateFunc
        self._equalityFunc      = equalityFunc
        self._allowInvalid      = allowInvalid
        self._defaultAttributes = atts


    def __copy__(self):
        """Returns a copy of this :class:`PropertyBase` instance.

        NOTE: This has not been rigorously tested.
        """
        newProp = type(self)()
        newProp.__dict__.update(self.__dict__)

        # This is the critical step - make sure
        # that the class labels from this instance
        # are not copied across to the new instance
        newProp._label = {}

        # Give the new object an independent
        # defaultAttributes dictionary
        newProp._defaultAttributes = dict(newProp._defaultAttributes)

        return newProp


    def _setLabel(self, cls, label):
        """Sets the property label for the given class. A :exc:`RuntimeError`
        is raised if a label already exists for the given class.
        """

        if cls in self._label:
            raise RuntimeError('The {} instance assigned to {}.{} is '
                               'already present as {}.{}. A PropertyBase '
                               'instance may only be present on a class '
                               'once. Declare a new instance.'.format(
                                   self.__class__.__name__,
                                   cls.__name__,
                                   label,
                                   cls.__name__,
                                   self._label[cls]))

        self._label[cls] = label


    def getLabel(self, instance):
        """Returns the property label for the given instance (more
        specifically, for the class of the given instance), or ``None``
        if no such label has been defined.
        """
        if instance is None: return None
        return self._label.get(type(instance), None)


    def enable(self, instance):
        """Enables this property for the given :class:`HasProperties`
        instance.

        See the :meth:`disable` method for more details.
        """
        propVal = self.getPropVal(instance)
        propVal.setAttribute('enabled', True)


    def disable(self, instance):
        """Disables this property for the given :class:`HasProperties`
        instance.

        An attempt to set the value of a disabled property will result in a
        :class:`DisabledError`. This behaviour can be circumvented by dealing
        directly with the underlying :class:`.PropertyValue` object.

        Changes to the enabled state of a property may be detected by
        registering an attribute listener (see
        :meth:`.PropertyValue.addAttributeListener`) and listening for changes
        to the ``enabled`` attribute.
        """
        propVal = self.getPropVal(instance)
        propVal.setAttribute('enabled', False)


    def isEnabled(self, instance):
        """Returns ``True`` if this property is enabled for the given
        :class:`HasProperties` instance, ``False`` otherwise.

        See the :meth:`disable` method for more details.
        """
        propVal = self.getPropVal(instance)
        return propVal.getAttribute('enabled')


    def addListener(self, instance, *args, **kwargs):
        """Register a listener with the ``PropertyValue`` object managed
        by this property. See :meth:`.PropertyValue.addListener`.

        :param instance:  The ``HasProperties`` instance on which the
                          listener is to be registered.
        """
        self._getInstanceData(instance).propVal.addListener(*args, **kwargs)


    def removeListener(self, instance, name):
        """De-register the named listener from the ``PropertyValue`` object
        managed by this property.
        """
        instData = self._getInstanceData(instance)

        if instData is None: return
        else:                instData.propVal.removeListener(name)


    def getConstraint(self, instance, constraint):
        """See :meth:`getAttribute`. """
        warnings.warn('getConstraint is deprecated - use getAttribute instead',
                      category=DeprecationWarning,
                      stacklevel=2)
        return self.getAttribute(instance, constraint)


    def setConstraint(self, instance, constraint, value):
        """See :meth:`setAttribute`. """
        warnings.warn('setConstraint is deprecated - use setAttribute instead',
                      category=DeprecationWarning,
                      stacklevel=2)
        return self.setAttribute(instance, constraint, value)


    def getAttribute(self, instance, att, *arg):
        """Returns the value of the named attribute for the specified
        ``HasProperties`` instance, or the default attribute value if
        instance is ``None``. See :meth:`.PropertiesValue.getAttribute`.
        """
        instData  = self._getInstanceData(instance)
        nodefault = len(arg) == 0

        if instData is None:
            if nodefault: return self._defaultAttributes[att]
            else:         return self._defaultAttributes.get(att, arg[0])
        else:
            return instData.propVal.getAttribute(att, *arg)


    def setAttribute(self, instance, att, value):
        """Sets the value of the named attribute for the specified
        ``HasProperties`` instance,or  the default value if instance
        is ``None``. See :meth:`.PropertiesValue.setAttribute`.
        """

        instData = self._getInstanceData(instance)
        oldVal   = self.getAttribute(instance, att, None)

        if value == oldVal: return

        log.debug('Changing {} attribute on {}: {} = {}'.format(
            ''        if instance is None else self.getLabel(instance),
            'default' if instance is None else 'instance',
            att,
            value))

        if instData is None: self._defaultAttributes[att] = value
        else:                instData.propVal.setAttribute(att, value)


    def getPropVal(self, instance):
        """Return the :class:`.PropertyValue` instance(s) for this property,
        associated with the given ``HasProperties`` instance, or ``None``
        if there is no value for the given instance.
        """
        instData = self._getInstanceData(instance)
        if instData is None: return None
        return instData.propVal


    def _getInstanceData(self, instance):
        """Returns the :class:`_InstanceData` object for the given
        ``HasProperties`` instance, or ``None`` if there is no
        ``_InstanceData`` for the given instance. An ``_InstanceData``
        object, which provides a binding between a ``PropertyBase``
        object and a ``HasProperties`` instance, is created by that
        ``HasProperties`` instance when it is created (see
        :meth:`HasProperties.__new__`).
        """
        if instance is None: return None
        return instance.__dict__.get(self.getLabel(instance), None)


    def _makePropVal(self, instance):
        """Creates and returns a ``PropertyValue`` object for the given
        ``HasProperties`` instance.
        """
        default = self._defaultAttributes.get('default', None)
        return properties_value.PropertyValue(instance,
                                              name=self.getLabel(instance),
                                              value=default,
                                              castFunc=self.cast,
                                              validateFunc=self.validate,
                                              equalityFunc=self._equalityFunc,
                                              allowInvalid=self._allowInvalid,
                                              **self._defaultAttributes)


    def validate(self, instance, attributes, value):
        """Called when an attempt is made to set the property value on the
        given instance.

        Called by ``PropertyValue`` objects when their value changes. The sole
        purpose of this method is to determine whether a given value is valid
        or invalid; it should not do anything else. In particular, it should
        not modify any other property values on the instance, as bad things
        will probably happen.

        If the given value is invalid, subclass implementations should raise a
        :exc:`ValueError` containing a useful message as to why the value is
        invalid. Otherwise, they should not return any value.  The default
        implementation does nothing, unless a custom validate function, and/or
        ``required=True``, was passed to the constructor. If ``required`` is
        ``True``, and the value is ``None``, a :exc:`ValueError` is raised. If
        a custom validate function was set, it is called and, if it returns
        ``False``, a :exc:`ValueError` is raised. It may also raise a
        :exc:`ValueError` of its own for invalid values.

        Subclasses which override this method should therefore call this
        superclass implementation in addition to performing their own
        validation.

        :param instance:        The ``HasProperties`` instance which
                                owns this ``PropertyBase`` instance,
                                or ``None`` for an unbound property value.

        :param dict attributes: Attributes of the ``PropertyValue`` object,
                                which are used to store type-specific
                                attributes for ``PropertyBase`` subclasses.

        :param value:           The value to be validated.
        """

        # a value is required
        if (self._required is not None) and (value is None):

            # required may either be a boolean value
            if isinstance(self._required, bool):
                if self._required:
                    raise ValueError('A value is required')

            # or a function
            elif self._required(instance):
                raise ValueError('A value is required')

        # a custom validation function has been provided
        if self._validateFunc is not None:
            if not self._validateFunc(instance, attributes, value):
                raise ValueError('Value does not meet custom validation rules')


    def cast(self, instance, attributes, value):
        """This method is called when a value is assigned to this
        ``PropertyBase`` instance through a ``HasProperties`` attribute
        access. The default implementaton just returns the given value.
        Subclasses may override this method to perform any required implicit
        casting or conversion rules.
        """
        return value


    def revalidate(self, instance):
        """Forces validation of this property value, for the current instance.
        This will result in any registered listeners being notified, but only
        if the validity of the value has changed.
        """

        propVal = self.getPropVal(instance)
        propVal.revalidate()


    def __get__(self, instance, owner):
        """If called on the ``HasProperties`` class, and not on an instance,
        returns this ``PropertyBase`` object. Otherwise, returns the value
        contained in the ``PropertyValue`` object which is attached to the
        instance.
        """

        if instance is None:
            return self

        instData = self._getInstanceData(instance)
        return instData.propVal.get()


    def __set__(self, instance, value):
        """Set the value of this property, as attached to the given instance,
        to the given value.
        """

        propVal = self.getPropVal(instance)

        if not propVal.getAttribute('enabled'):
            raise DisabledError('Property {}.{} is disabled'.format(
                instance.__class__.__name__,
                self.getLabel(instance)))

        propVal.set(value)


class ListPropertyBase(PropertyBase):
    """A :class:`PropertyBase` for properties which encapsulate more than
    one value.
    """

    def __init__(self, listType, **kwargs):
        """Define a ``ListPropertyBase`` property.

        :param listType: An unbound ``PropertyBase`` instance, defining
                         the type of value allowed in the list. This is
                         optional; if not provided, values of any type will be
                         allowed in the list, but no validation or casting
                         will be performed.
        """
        PropertyBase.__init__(self, **kwargs)
        self._listType = listType


    def getListType(self):
        """Returns a reference to the ``PropertyBase`` instance which defines
        the value types allowsed in this ``ListPropertyBase``. This may be
        ``None``.
        """
        return self._listType


    def _makePropVal(self, instance):
        """Creates and returns a :class:`.PropertyValueList` object to be
        associated with the given ``HasProperties`` instance.
        """

        if self._listType is not None:
            itemCastFunc     = self._listType.cast
            itemValidateFunc = self._listType.validate
            itemEqualityFunc = self._listType._equalityFunc
            itemAllowInvalid = self._listType._allowInvalid
            itemAttributes   = self._listType._defaultAttributes
        else:
            itemCastFunc     = None
            itemValidateFunc = None
            itemEqualityFunc = None
            itemAllowInvalid = True
            itemAttributes   = None

        default = self._defaultAttributes.get('default', None)

        return properties_value.PropertyValueList(
            instance,
            name=self.getLabel(instance),
            values=default,
            itemCastFunc=itemCastFunc,
            itemValidateFunc=itemValidateFunc,
            itemEqualityFunc=itemEqualityFunc,
            listValidateFunc=self.validate,
            itemAllowInvalid=itemAllowInvalid,
            listAttributes=self._defaultAttributes,
            itemAttributes=itemAttributes)


    def enableItem(self, instance, index):
        """Enables the item at the given ``index``.  See :meth:`disableItem`.
        """
        self.getPropValList(instance)[index].setAttribute('enabled', True)


    def disableItem(self, instance, index):
        """Disables the item at the given ``index``.  See
        :meth:`PropertyBase.disable`.

        .. note:: ``ListPropertyBase`` items cannot actually be disabled
                  at this point in time - all this method does is set the
                  ``'enabled'`` attribute of the item to ``False``.
        """
        self.getPropValList(instance)[index].setAttribute('enabled', False)


    def getPropValList(self, instance):
        """Returns the list of ``PropertyValue`` objects which represent the
        items stored in this list.

        Note that this is a list of ``PropertyValue`` instances; it is not the
        ``PropertyValueList`` instance. The latter can be accessed through
        the owning ``HasProperties`` instance with a simple attribute access.
        """
        propVal = self.getPropVal(instance)
        if propVal is not None: return propVal.getPropertyValueList()
        else:                   return None


class PropertyOwner(type):
    """Deprecated. """
    def __new__(cls, name, bases, attrs):
        warnings.warn('PropertyOwner is deprecated and is no longer used',
                      category=DeprecationWarning,
                      stacklevel=2)
        return super(PropertyOwner, cls).__new__(cls, name, bases, attrs)


class HasProperties:
    """Base class for classes which contain ``PropertyBase`` instances.  All
    classes which contain ``PropertyBase`` objects must subclass this
    class.

    .. note:: ``HasProperties`` is also available via an alias called
              :attr:`HasProps`.
    """


    def __new__(cls, *args, **kwargs):
        """Here we create a new ``HasProperties`` instance, and loop
        through all of its ``PropertyBase`` properties to ensure that
        they are initialised.
        """

        instance = super(HasProperties, cls).__new__(cls)

        # By default, when a property changes,
        # all other properties are not validated.
        # This behaviour can be changed by passing
        # validateOnChange=True to __init__.
        instance.__validateOnChange = False

        # Helper to return *all* attributes of a
        # class, including those of its base classes
        def allAttrs(cls):
            atts = list(cls.__dict__.items())
            if hasattr(cls, '__bases__'):
                for base in cls.__bases__:
                    atts += list(allAttrs(base))
            return atts

        # Add each class level PropertyBase
        # object as a property of the new
        # HasProperties instance
        for propName, propObj in allAttrs(cls):

            if not isinstance(propObj, PropertyBase):
                continue

            # A PropertyBase instance maintains labels
            # for each class, so this will only need to
            # be done for the first instance that gets
            # created.
            if propObj.getLabel(instance) is None:
                propObj._setLabel(cls, propName)

            instance.addProperty(propName, propObj)

        return instance


    def __init__(self, validateOnChange=False, **kwargs):
        """Create a ``HasProperties`` instance.

        If no arguments need to be passed in, ``HasProperties.__init__`` does
        not need to be called.

        :arg validateOnChange: Defaults to ``False``. If set to ``True``,
                               whenever any property value is changed,
                               the value of *every* property is re-validated.
                               This functionality is accomplished by using
                               the *preNotify* listener on all
                               ``PropertyValue`` instances - see the
                               :meth:`.PropertyValue.setPreNotifyFunction`
                               method.

        :arg kwargs:           All other arguments are assumed to be
                               ``name=value`` pairs, containing initial values
                               for the properties defined on this
                               ``HasProperties`` instance.

        .. note:: The ``validateOnChange`` argument warrants some explanation.

           The point of validating all other properties when one property
           changes is to handle the scenario where the validity of one
           property is dependent upon the values of other properties.

           Currently, the only option is to enable this globally;
           i.e. whenever the value of any property changes, all other
           properties are validated.

           At some stage, I may allow more fine grained control;
           e.g. validation could only occur when specific properties change,
           and/or only specific properties are validated. This should be
           fairly straightforward - we could just maintain a dict of
           ``{propName : [propNames ..]}`` mappings, where the key is the name
           of a property that should trigger validation, and the value is a
           list of properties that need to be validated when that property
           changes.
        """

        self.__validateOnChange = validateOnChange

        # The prenotify function is added to new properties
        # in the addProperty method, but not for properties
        # defined at the class level (because the
        # __validateOnChange attribute is initially set to
        # false in __new__). So here we make sure that the
        # prenotify is set if needed.

        if validateOnChange:
            propNames, props = self.getAllProperties()

            for prop, propName in zip(propNames, props):
                propVal = prop.getPropVal(self)
                propVal.setPreNotifyFunction(self.__valueChanged)

        # Initial values
        for name, value in kwargs.items():
            setattr(self, name, value)


    def __copy__(self):
        """Default copy operator.

        Creates a new instance of this type, and copies all property values
        across.

        If a no-arguments constructor is not available, an error will
        be raised.

        Subclasses which require arguments on initialisation, or which have
        more complex copy semantics, will need to implement their own
        ``__copy__`` operator if this one does not suffice.
        """

        copy = type(self)()

        # TODO Is this going to crash for List properties?
        #      If it does, make it not crash.
        for propName in self.getAllProperties()[0]:
            setattr(copy, propName, getattr(self, propName))

        return copy


    def bindProps(self, *args, **kwargs):
        """See :func:`.bindable.bindProps`. """
        bindable.bindProps(self, *args, **kwargs)


    def unbindProps(self, *args, **kwargs):
        """See :func:`.bindable.unbindProps`. """
        bindable.unbindProps(self, *args, **kwargs)


    def bind(self, *args, **kwargs):
        """Alias for :meth:`bindProps`. """
        self.bindProps(*args, **kwargs)


    def unbind(self, *args, **kwargs):
        """Alias for :meth:`unbindProps`. """
        self.unbindProps(*args, **kwargs)


    def isBound(self, *args, **kwargs):
        """See :func:`.bindable.isBound`. """
        bindable.isBound(self, *args, **kwargs)


    def addProperty(self, propName, propObj):
        """Add the given `PropertyBase`` instance as an attribute of this
        ``HasProperties`` instance. """
        if not isinstance(propObj, PropertyBase):
            raise ValueError('propObj must be a PropertyBase instance')

        # If this property does not exist on the class,
        # add it. This is a bit hacky, as the labels
        # for all the properties that exist in the class
        # definition are handled by the metaclass. What's
        # stopping me from throwing out the metaclass,
        # and doing everything in HasProps.__new__, and
        # this method? Related - is there a reason why
        # PropertyBase labels are tied to the HasProps
        # class, rather than to the instance?
        if not hasattr(self.__class__, propName):
            setattr(          self.__class__, propName, propObj)
            propObj._setLabel(self.__class__, propName)

        # Create a PropertyValue and an _InstanceData
        # object, which bind the PropertyBase object
        # to this HasProperties instance.
        propVal = propObj._makePropVal(self)
        instData = _InstanceData(self, propVal)

        log.debug('Adding property to {}.{} [{}] ({})'.format(
            self.__class__.__name__,
            propName,
            id(self),
            propObj.__class__.__name__))

        # Store the _InstanceData object
        # on this instance itself
        self.__dict__[propName] = instData

        # validate other properties when
        # this property changes - does
        # nothing if validation is enabled
        if self.__validateOnChange:
            propVal.setPreNotifyFunction(self.__valueChanged)


    def __valueChanged(self, ctx, value, valid, name):
        """This method is only called if ``validateOnChange`` was set
        to true in :meth:`__init__`. It is registered as the ``preNotify``
        listener on all ``PropertyValue`` instances. See the note in
        :meth:`__init__`.
        """

        if not self.__validateOnChange:
            return

        # Force validation for all other properties of the instance, and
        # notification of their registered listeners, This is done because the
        # validity of some properties may be dependent upon the values of this
        # one. So when the value of this property changes, it may have changed
        # the validity of another property, meaning that the listeners of the
        # latter property need to be notified of this change in validity.

        log.debug('Revalidating all instance properties '
                  '(due to {} change)'.format(name))

        propNames, props = self.getAllProperties()
        for propName, prop in zip(propNames, props):
            if propName is not name:
                prop.revalidate(self)


    @classmethod
    def getAllProperties(cls):
        """Returns two lists, the first containing the names of all properties
        of this object, and the second containing the corresponding
        ``PropertyBase`` objects.

        Properties which have a name beginning with an underscore are not
        returned by this method
        """

        propNames = []
        props     = []

        for attName in dir(cls):

            att = getattr(cls, attName)

            if isinstance(att, PropertyBase) and (not attName.startswith('_')):
                propNames.append(attName)
                props    .append(att)

        return propNames, props


    @classmethod
    def getProp(cls, propName):
        """Return the ``PropertyBase`` object for the given property."""
        return getattr(cls, propName)


    def getPropVal(self, propName):
        """Return the ``PropertyValue`` object(s) for the given property.
        """
        return self.getProp(propName).getPropVal(self)


    def getLastValue(self, propName):
        """Returns the most recent value of the specified property before its
        current one.

        See the :meth:`.PropertyValue.getLast` method.
        """
        return self.getPropVal(propName).getLast()


    def serialise(self, propName):
        """Returns a string containing the value of the named property,
        serialsied via the :mod:`.serialise` module.
        """
        return serialise.serialise(self, propName)


    def deserialise(self, propName, value):
        """Deserialises the given value (assumed to have been serialised
        via the :mod:`.serialise` module), and sets the named property
        to the deserialised value.
        """
        value = serialise.deserialise(self, propName, value)
        setattr(self, propName, value)


    def enableNotification(self, propName, *args, **kwargs):
        """Enables notification of listeners on the given property.

        See the :meth:`.PropertyValue.enableNotification` method.
        """
        self.getPropVal(propName).enableNotification(*args, **kwargs)


    def disableNotification(self, propName, *args, **kwargs):
        """Disables notification of listeners on the given property.

        See the :meth:`.PropertyValue.disableNotification` method.
        """
        self.getPropVal(propName).disableNotification(*args, **kwargs)


    def getNotificationState(self, propName):
        """Returns the notification state of the given property.

        See the :meth:`.PropertyValue.getNotificationState` method.
        """
        return self.getPropVal(propName).getNotificationState()


    def setNotificationState(self, propName, value):
        """Sets the notification state of the given property.

        See the :meth:`.PropertyValue.setNotificationState` method.
        """
        self.getPropVal(propName).setNotificationState(value)


    def enableAllNotification(self):
        """Enables notification of listeners on all properties."""
        propNames, props = self.getAllProperties()
        for propName in propNames:
            self.enableNotification(propName)


    def disableAllNotification(self):
        """Disables notification of listeners on all properties."""
        propNames, props = self.getAllProperties()
        for propName in propNames:
            self.disableNotification(propName)


    def enableProperty(self, propName, index=None):
        """Enables the given property.

        If an ``index`` is provided, it is assumed that the property is a
        list property (a :class:`ListPropertyBase`).

        See :meth:`PropertyBase.enable` and
        :meth:`ListPropertyBase.enableItem`.
        """
        if index is not None: self.getProp(propName).enableItem(self, index)
        else:                 self.getProp(propName).enable(    self)


    def disableProperty(self, propName, index=None):
        """Disables the given property - see :meth:`PropertyBase.disable`.

        If an ``index`` is provided, it is assumed that the property is a
        list property (a :class:`ListPropertyBase`).

        See :meth:`PropertyBase.disable` and
        :meth:`ListPropertyBase.disableItem`.
        """
        if index is not None: self.getProp(propName).disableItem(self, index)
        else:                 self.getProp(propName).disable(    self)


    def propertyIsEnabled(self, propName):
        """Returns the enabled state of the given property - see
        :meth:`PropertyBase.isEnabled`.
        """
        return self.getProp(propName).isEnabled(self)


    def propNotify(self, propName):
        """Force notification of listeners on the given property. This will
        have no effect if notification for the property is disabled.

        See the :meth:`.PropertyValue.propNotify` method.
        """
        self.getPropVal(propName).propNotify()


    def getConstraint(self, propName, constraint):
        """See :meth:`getAttribute`. """
        warnings.warn('getConstraint is deprecated - use getAttribute instead',
                      category=DeprecationWarning,
                      stacklevel=2)
        return self.getAttribute(propName, constraint)


    def setConstraint(self, propName, constraint, value):
        """See :meth:`setAttribute`. """
        warnings.warn('setConstraint is deprecated - use setAttribute instead',
                      category=DeprecationWarning,
                      stacklevel=2)
        return self.setAttribute(propName, constraint, value)


    def getAttribute(self, propName, *args):
        """Convenience method, returns the value of the named attributes for
        the named property. See :meth:`PropertyBase.getAttribute`.
        """
        return self.getProp(propName).getAttribute(self, *args)


    def setAttribute(self, propName, *args):
        """Convenience method, sets the value of the named attribute for
        the named property. See :meth:`PropertyBase.setAttribute`.
        """
        return self.getProp(propName).setAttribute(self, *args)


    def setatt(self, *args, **kwargs):
        """Alias for :meth:`setAttribute`. """
        return self.setAttribute(*args, **kwargs)


    def getatt(self, *args, **kwargs):
        """Alias for :meth:`getAttribute`. """
        return self.getAttribute(*args, **kwargs)


    def listen(self, *args, **kwargs):
        """Alias for :meth:`addListener`. """
        self.addListener(*args, **kwargs)


    def ilisten(self, *args, **kwargs):
        """Calls ``addListener(immediate=True)``. """
        self.addListener(*args, immediate=True, **kwargs)


    def remove(self, *args, **kwargs):
        """Alias for :meth:`removeListener`. """
        self.removeListener(*args, **kwargs)


    def addListener(self, propName, *args, **kwargs):
        """Convenience method, adds the specified listener to the specified
        property. See :meth:`PropertyValue.addListener`.
        """
        self.getPropVal(propName).addListener(*args, **kwargs)


    def removeListener(self, propName, listenerName):
        """Convenience method, removes the specified listener from the
        specified property. See :meth:`PropertyValue.removeListener`.
        """
        self.getPropVal(propName).removeListener(listenerName)


    def enableListener(self, propName, name):
        """(Re-)Enables the listener on the specified property with the
        specified ``name``.
        """
        self.getPropVal(propName).enableListener(name)


    def disableListener(self, propName, name):
        """Disables the listener on the specified property with the specified
        ``name``, but does not remove it from the list of listeners.
        """
        self.getPropVal(propName).disableListener(name)


    def getListenerState(self, propName, name):
        """See :meth:`.PropertyValue.getListenerState`. """
        return self.getPropVal(propName).getListenerState(name)


    def setListenerState(self, propName, name, state):
        """See :meth:`.PropertyValue.setListenerState`. """
        self.getPropVal(propName).setListenerState(name, state)


    def hasListener(self, propName, name):
        """Returns ``True`` if a listener is registered on the given property,
        ``False`` otherwise.
        """
        return self.getPropVal(propName).hasListener(name)


    def addGlobalListener(self, *args, **kwargs):
        """Registers the given listener so that it will be notified of
        changes to any of the properties of this HasProperties instance.
        """
        propNames, props = self.getAllProperties()
        for propName in propNames:
            self.getPropVal(propName).addListener(*args, **kwargs)


    def removeGlobalListener(self, listenerName):
        """De-registers the specified global listener (see
        :meth:`addGlobalListener`).
        """
        propNames, props = self.getAllProperties()
        for propName in propNames:
            self.getPropVal(propName).removeListener(listenerName)


    def isValid(self, propName):
        """Returns ``True`` if the current value of the specified property is
        valid, ``False`` otherwise.
        """

        prop    = self.getProp(propName)
        propVal = prop.getPropVal(self)

        try: prop.validate(self, propVal.getAttributes(), propVal.get())
        except ValueError: return False

        return True


    def validateAll(self):
        """Validates all of the properties of this :class:`HasProperties`
        object.  A list of tuples is returned, with each tuple containing
        a property name, and an associated error string. The error string
        is a message about the property which failed validation. If all
        property values are valid, the returned list will be empty.
        """

        names, props = self.getAllProperties()

        errors = []

        for name, prop in zip(names, props):

            propVal = prop.getPropVal(self)

            try:
                prop.validate(self, propVal.getAttributes(), propVal.get())

            except ValueError as e:
                errors.append((name, e))

        return errors


    def __str__(self):
        """Returns a multi-line string containing the names and values of
        all the properties of this object.
        """

        clsname = self.__class__.__name__

        propNames, props = self.getAllProperties()

        if len(propNames) == 0:
            return clsname

        propVals = ['{}'.format(getattr(self, propName))
                    for propName in propNames]

        maxNameLength = max(map(len, propNames))

        lines = [clsname]

        for propName, propVal in zip(propNames, propVals):
            fmtStr = '  {:>' + str(maxNameLength) + '} = {}'
            lines.append(fmtStr.format(propName, propVal))

        return '\n'.join(lines)


HasProps = HasProperties
"""``HasProps`` is simply an alias for :class:`HasProperties`. """


Props = HasProperties
"""``Props`` is simply an alias for :class:`HasProperties`. """



#!/usr/bin/env python
#
# test_callqueue.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import time
import threading
import logging
import numpy as np

import fsleyes_props.callqueue as callqueue


def test_call():

    ncalls = 10
    called = [0] * ncalls
    calls  = []

    for i in range(ncalls):
        def func(ii=i):
            called[ii] = ii

        calls.append(func)

    q = callqueue.CallQueue()

    for i, c in enumerate(calls):
        q.call(c, 'Call {}'.format(i))

    assert called == list(range(ncalls))


# callqueu has some internal functions that are
# only called if logging level >= DEBUG
def test_debug():

    def func():
        pass

    logging.basicConfig()
    log = logging.getLogger('fsleyes_props.callqueue')
    log.setLevel(logging.DEBUG)

    q = callqueue.CallQueue()
    q.call(func, 'func', ())

    log.setLevel(logging.WARNING)


def test_callAll():

    ncalls = 10
    called = [0] * ncalls
    calls  = []

    for i in range(ncalls):
        def func(ii=i):
            called[ii] = ii

        calls.append(func)

    q     = callqueue.CallQueue()
    calls = [(c, 'Call {}'.format(i), (), {}) for i, c in enumerate(calls)]

    q.callAll(calls)

    assert called == list(range(ncalls))


def test_call_raise():

    def badfunc():
        raise Exception('Call that was supposed to crash crashed!')

    # Not crashing means the test passes
    q = callqueue.CallQueue()
    q.call(badfunc, 'Bad function', ())


def test_dequeue():

    q      = callqueue.CallQueue()
    called = [False, False]

    def func0():
        q.dequeue('func1')
        called[0] = True

    def func1():
        called[1] = True

    calls = [(func0, 'func0', (), {}), (func1, 'func1', (), {})]

    q.callAll(calls)

    assert     called[0]
    assert not called[1]


def test_call_recursive():

    q         = callqueue.CallQueue()
    callOrder = []

    def func0():
        callOrder.append(0)
        q.call(func1, 'func1')

    def func1():
        callOrder.append(1)

    def func2():
        callOrder.append(2)

    calls = [(func0, 'func0', (), {}), (func2, 'func2', (), {})]

    q.callAll(calls)

    assert callOrder == [0, 2, 1]


def test_call_threaded():

    q = callqueue.CallQueue()

    nthreads = 1000
    calltime = {}
    flags    = [threading.Event() for i in range(nthreads)]

    def func(delay):
        calltime[delay] = time.time()
        time.sleep(np.random.random() * 0.1)

    def threadfunc(delay):
        if delay > 0:
            flags[delay - 1].wait()
        q.call(func, 'func_{}'.format(delay), delay)
        flags[delay].set()

    delays  = np.arange(20)
    threads = [threading.Thread(target=threadfunc, args=(d,)) for d in delays]

    for t in threads:
        t.start()

    for t in threads:
        t.join()

    calltimes = [calltime[d] for d in delays]
    assert calltimes == sorted(calltimes)


def test_dequeue_threaded():

    q = callqueue.CallQueue()

    called = {}

    def func(delay):
        called[delay] = True
        time.sleep(delay / 5.0)

    def threadfunc():
        calls = [(func, 'slowfunc', (5,), {}),
                 (func, 'fastfunc', (1,), {})]
        q.callAll(calls)

    t = threading.Thread(target=threadfunc)

    t.start()
    time.sleep(0.1)

    q.dequeue('fastfunc')
    t.join()

    assert     called[5]
    assert not called.get(1, False)


def test_skipDuplicates():   _test_skipDuplicates(True)
def test_noSkipDuplicates(): _test_skipDuplicates(False)
def _test_skipDuplicates(skip):


    nthreads = 5
    q        = callqueue.CallQueue(skipDuplicates=skip)
    ncalls   = [0]

    def func():
        ncalls[0] += 1
        time.sleep(0.5)

    def threadfunc(name):
        q.call(func, name)

    threads = [threading.Thread(target=threadfunc, args=('func', ))
               for i in range(nthreads)]
    blockt  = threading.Thread(target=threadfunc, args=('block',))

    blockt.start()
    for t in threads:
        t.start()

    blockt.join()
    for t in threads:
        t.join()

    if skip: assert ncalls[0] == 2
    else:    assert ncalls[0] == nthreads + 1


def test_dequeue_noSkipDuplicates():

    nthreads = 5

    q = callqueue.CallQueue(skipDuplicates=True)

    ncalls = [0]

    def func():
        ncalls[0] += 1
        time.sleep(0.5)

    def threadfunc(name):
        q.call(func, name)

    threads = [threading.Thread(target=threadfunc, args=('func', ))
               for i in range(nthreads)]
    blockt  = threading.Thread(target=threadfunc, args=('block',))

    blockt.start()
    time.sleep(0.1)
    for t in threads:
        t.start()
    q.dequeue('func')

    blockt.join()
    for t in threads:
        t.join()

    assert ncalls[0] == 1


def test_hold():

    q      = callqueue.CallQueue()
    called = {}

    def func(name):
        called[name] = called.get(name, 0) + 1

    q.call(func, 'one', 'one')
    q.hold()
    q.call(func, 'two',   'two')
    q.call(func, 'three', 'three')
    q.call(func, 'four',  'four')

    # clearheld should only
    # return functions when
    # the queue is not being
    # held
    assert q.clearHeld() == []

    assert called['one'] == 1
    assert 'two'   not in called
    assert 'three' not in called
    assert 'four'  not in called

    # Try and dequeue a held function
    q.dequeue('two')
    q.release()
    q.call(func, 'five', 'five')

    assert called['five'] == 1
    assert 'two'   not in called
    assert 'three' not in called
    assert 'four'  not in called

    q.callAll(q.clearHeld())

    assert called['one']   == 1
    assert 'two'  not in called
    assert called['three'] == 1
    assert called['four']  == 1
    assert called['five']  == 1

def test_hold_threaded():

    # TODO
    assert True



#!/usr/bin/env python
#
# test_widget_boolean.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import os.path as op
import wx

import fsleyes_props                as props
import fsleyes_widgets.bitmaptoggle as bmptoggle
import fsleyes_widgets.bitmapradio  as bmpradio

from . import run_with_wx, addall, realYield


datadir = op.join(op.dirname(__file__), 'testdata')


def setup_module():
    props.initGUI()


class MyObj(props.HasProperties):
    mybool = props.Boolean()


def  test_widget_boolean(): run_with_wx(_test_widget_boolean)
def _test_widget_boolean(parent):

    trueicon  = op.join(datadir, 'true.png')
    falseicon = op.join(datadir, 'false.png')

    obj = MyObj()

    # ch
    myboolcb  = props.makeWidget(parent, obj, 'mybool')
    mybooltog = props.makeWidget(parent, obj, 'mybool',
                                 icon=[trueicon, falseicon])
    myboolrad = props.makeWidget(parent, obj, 'mybool',
                                 icon=[trueicon, falseicon],
                                 toggle=False)

    assert isinstance(myboolcb,  wx.CheckBox)
    assert isinstance(mybooltog, bmptoggle.BitmapToggleButton)
    assert isinstance(myboolrad, bmpradio .BitmapRadioBox)

    addall(parent, (myboolcb, mybooltog, myboolrad))

    obj.mybool = False

    assert not myboolcb .GetValue()
    assert not mybooltog.GetValue()
    assert myboolrad.GetSelection() == 1
    obj.mybool = True

    assert myboolcb .GetValue()
    assert mybooltog.GetValue()
    assert myboolrad.GetSelection() == 0

    myboolcb.SetValue(False)
    wx.PostEvent(myboolcb, wx.CommandEvent(wx.EVT_CHECKBOX.evtType[0]))
    realYield()
    assert not obj.mybool
    assert not myboolcb .GetValue()
    assert not mybooltog.GetValue()
    assert myboolrad.GetSelection() == 1

    myboolcb.SetValue(True)
    wx.PostEvent(myboolcb, wx.CommandEvent(wx.EVT_CHECKBOX.evtType[0]))
    realYield()
    assert obj.mybool
    assert myboolcb .GetValue()
    assert mybooltog.GetValue()
    assert myboolrad.GetSelection() == 0
